'''
UI create layout
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import warnings
from PyQt6.QtWidgets import (QMainWindow,
                             QGridLayout,
                             QHBoxLayout,
                             QVBoxLayout,
                             QTabWidget,
                             QWidget)
from IQM_Vis.UI import utils

# sub class used by IQM_Vis.main.make_app to initialise layout of the UI
# uses widgets from IQM_Vis.widgets.app_widgets
# class layout(QMainWindow):
class layout(QMainWindow):
    def __init__(self, **kwargs):
        super().__init__()

    def init_layout(self):
        self.tabs = {}
        self._init_generic_layout()
        self.main_widget = QWidget()
        self.main_widget.setLayout(self.main_layout)
        self.setCentralWidget(self.main_widget)
        self.show()
        return self

    def _init_generic_layout(self):
        '''
        place all the widgets in the window
        '''
        # sizes
        im_width = 10
        im_height = 4
        # save window tabs to self
        self.tabs = {}

        '''image rows displaying image, metric graphs and metric images'''
        image_layouts = QVBoxLayout()
        metric_layouts = QVBoxLayout()
        graph_layouts = QVBoxLayout()
        for i in self.widget_row:
            '''image and transformed image'''
            image_layout = QHBoxLayout()
            for image_name in self.widget_row[i]['images']:
                single_image = QVBoxLayout()
                single_image.addWidget(self.widget_row[i]['images'][image_name]['label'])
                single_image.addWidget(self.widget_row[i]['images'][image_name]['data'])
                image_layout.addLayout(single_image)
            image_layouts.addLayout(image_layout)
            image_layouts.addStretch()
            '''metric images'''
            metric_layout = QHBoxLayout()
            for metric_name in self.widget_row[i]['metric_images']:
                single_metric = QVBoxLayout()
                single_metric.addWidget(self.widget_row[i]['metric_images'][metric_name]['label'])
                single_metric.addWidget(self.widget_row[i]['metric_images'][metric_name]['data'])
                single_metric.addStretch()
                metric_layout.addLayout(single_metric)
                metric_layout.addStretch()
            metric_layouts.addLayout(metric_layout)
            metric_layouts.addStretch()

            '''graphs'''
            self.tabs['graph'] = QTabWidget()
            # '''metrics graphs'''
            metric_bar = QVBoxLayout()
            metric_bar.addWidget(self.widget_row[i]['metrics']['info']['label'])
            metric_bar.addWidget(self.widget_row[i]['metrics']['info']['data'])
            utils.add_layout_to_tab(self.tabs['graph'], metric_bar, 'Metrics')
            if 'avg' in self.widget_row[i]['metrics'].keys():
                avg_graph = QVBoxLayout()
                avg_graph.addWidget(self.widget_row[i]['metrics']['avg']['label'])
                graph = QGridLayout()
                graph.addWidget(self.widget_row[i]['metrics']['avg']['data'], 0, 0, im_height, im_width)
                avg_graph.addLayout(graph)   # need for matplotlib? - test this...   (grid)
                graph_save = QHBoxLayout()
                graph_save.addWidget(self.widget_row[i]['metrics']['avg']['save_button'])
                avg_graph.addLayout(graph_save)
                utils.add_layout_to_tab(self.tabs['graph'], avg_graph, 'Radar')
            if 'range' in self.widget_row[i]['metrics'].keys():
                range_graph = QVBoxLayout()
                range_graph.addWidget(self.widget_row[i]['metrics']['range']['label'])
                graph = QGridLayout()
                graph.addWidget(self.widget_row[i]['metrics']['range']['data'], 0, 0, im_height, im_width)
                range_graph.addLayout(graph)  # need for matplotlib? - test this...    (grid)
                '''graph controls''' # will add to last one since there is one widget to control all
                graph_controls = QHBoxLayout()
                graph_controls.addWidget(self.widget_controls['button']['prev_metric_graph'])
                graph_controls.addWidget(self.widget_controls['button']['next_metric_graph'])
                range_graph.addLayout(graph_controls)
                graph_save = QHBoxLayout()
                graph_save.addWidget(self.widget_row[i]['metrics']['range']['save_button'])
                range_graph.addLayout(graph_save)
                utils.add_layout_to_tab(self.tabs['graph'], range_graph, 'Range')
            if 'correlation' in self.widget_row[i]['metrics'].keys():
                correlation_graph = QVBoxLayout()
                correlation_graph.addWidget(self.widget_row[i]['metrics']['correlation']['label'])
                graph = QGridLayout()
                graph.addWidget(self.widget_row[i]['metrics']['correlation']['data'], 0, 0, im_height, im_width)
                correlation_graph.addLayout(graph)  # need for matplotlib? - test this...    (grid)
                '''graph controls''' # will add to last one since there is one widget to control all
                graph_controls = QHBoxLayout()
                graph_controls.addWidget(self.widget_controls['button']['prev_correlation_graph'])
                graph_controls.addWidget(self.widget_controls['button']['next_correlation_graph'])
                correlation_graph.addLayout(graph_controls)
                '''load experiment'''
                correlation_graph.addWidget(self.widget_controls['button']['load_exp'])
                utils.add_layout_to_tab(self.tabs['graph'], correlation_graph, 'Correlation')
            if 'JND' in self.widget_row[i]['metrics'].keys():
                JND_graph = QVBoxLayout()
                JND_graph.addWidget(self.widget_row[i]['metrics']['JND']['label'])
                graph = QGridLayout()
                graph.addWidget(self.widget_row[i]['metrics']['JND']['data'], 0, 0, im_height, im_width)
                JND_graph.addLayout(graph)  # need for matplotlib? - test this...    (grid)
                '''graph controls''' # will add to last one since there is one widget to control all
                graph_controls = QHBoxLayout()
                JND_graph.addLayout(graph_controls)
                '''load experiment'''
                JND_graph.addWidget(self.widget_controls['button']['load_exp_JND'])
                utils.add_layout_to_tab(self.tabs['graph'], JND_graph, 'JND')

            graph_layouts.addWidget(self.tabs['graph'])
            graph_layouts.addStretch()

        '''dataset controls'''
        # prev/next buttons
        dataset_layout = QHBoxLayout()
        if self.dataset:
            dataset_layout.addWidget(self.widget_controls['label']['data'])
            dataset_layout.addWidget(self.widget_controls['button']['prev_data'])
            for im in range(self.num_images_scroll_show):
                dataset_layout.addWidget(self.widget_controls['images'][im])
            dataset_layout.addWidget(self.widget_controls['button']['next_data'])
            dataset_layout.addWidget(self.widget_controls['label']['data_num'])

        '''transform controls'''
        image_controls = QVBoxLayout()
        # transform sliders
        for key in self.sliders['transforms']:
            tran_layout_top = QHBoxLayout()
            tran_layout_top.addWidget(self.widget_controls['slider'][key]['label'])
            tran_layout_top.addWidget(self.widget_controls['slider'][key]['value'])
            tran_layout_bottom = QHBoxLayout()
            tran_layout_bottom.addWidget(self.sliders['transforms'][key]['min_edit'])
            tran_layout_bottom.addWidget(self.widget_controls['slider'][key]['data'])
            tran_layout_bottom.addWidget(self.sliders['transforms'][key]['max_edit'])
            trans_layout = QVBoxLayout()
            trans_layout.addLayout(tran_layout_top)
            trans_layout.addLayout(tran_layout_bottom)

            image_controls.addLayout(trans_layout)
            image_controls.addStretch()
        reset_button = QHBoxLayout()
        reset_button.addWidget(self.widget_controls['button']['transforms']['reset_sliders'])
        reset_button.addStretch()
        image_controls.addLayout(reset_button)

        '''metric_param controls'''
        metric_controls = QVBoxLayout()
        for key in self.sliders['metric_params']:
            inner_metric_layout = QHBoxLayout()
            inner_metric_layout.addWidget(self.widget_controls['slider'][key]['label'])
            inner_metric_layout.addWidget(self.widget_controls['slider'][key]['data'])
            inner_metric_layout.addWidget(self.widget_controls['slider'][key]['value'])
            metric_controls.addLayout(inner_metric_layout)
        reset_button = QHBoxLayout()
        reset_button.addWidget(self.widget_controls['button']['metric_params']['reset_sliders'])
        reset_button.addStretch()

        metric_controls.addLayout(reset_button)
        metric_controls.addStretch()

        '''settings controls'''
        settings_controls = QVBoxLayout()
        for key, item in self.widget_settings.items():
            inner_layout = QHBoxLayout()
            if isinstance(item, dict):
                inner_layout.addWidget(item['label'])
                inner_layout.addWidget(item['widget'])
            else:
                inner_layout.addWidget(item)
            inner_layout.addStretch()
            settings_controls.addLayout(inner_layout)
        settings_controls.addStretch()

        ''' experiment controls'''
        experiment_controls = QVBoxLayout()
        # tranformations
        for trans_name in self.widget_experiment_params:
            experiment_trans = QHBoxLayout()
            for _, widget in self.widget_experiment_params[trans_name].items():
                experiment_trans.addWidget(widget)
                experiment_trans.addStretch()
            experiment_controls.addLayout(experiment_trans)
        experiment_controls.addStretch()
        # Save folder
        save_button_exp = QHBoxLayout()
        save_button_exp.addWidget(self.widget_controls['label']['exp_change_save'])
        save_button_exp.addWidget(self.widget_controls['button']['exp_change_save'])
        save_button_exp.addStretch()
        experiment_controls.addLayout(save_button_exp)
        # Dataset Name
        dataset_name_exp = QHBoxLayout()
        dataset_name_exp.addWidget(
            self.widget_controls['label']['exp_change_dataset_name'])
        dataset_name_exp.addWidget(
            self.widget_controls['button']['exp_change_dataset_name'])
        dataset_name_exp.addStretch()
        experiment_controls.addLayout(dataset_name_exp)
        # Dataset Range nums
        dataset_range_exp = QHBoxLayout()
        dataset_range_exp.addWidget(
            self.widget_controls['label']['exp_im_nums'])
        dataset_range_exp.addWidget(
            self.widget_controls['label']['range_edit_lower'])
        dataset_range_exp.addWidget(
            self.widget_controls['label']['exp_im_nums_dash'])
        dataset_range_exp.addWidget(
            self.widget_controls['label']['range_edit_upper'])
        dataset_range_exp.addStretch()
        experiment_controls.addLayout(dataset_range_exp)

        # run experiment button
        experiment_button = QHBoxLayout()
        experiment_button.addWidget(self.widget_controls['button']['launch_exp_2AFC'])
        experiment_button.addWidget(self.widget_controls['button']['launch_exp_JND'])
        experiment_button.addStretch()
        experiment_controls.addLayout(experiment_button)

        '''export images'''
        export_controls = QVBoxLayout()
        for trans_name in self.widget_export:
            export_trans = QHBoxLayout()
            for _, widget in self.widget_export[trans_name].items():
                export_trans.addWidget(widget)
                export_trans.addStretch()
            export_controls.addLayout(export_trans)
        export_controls.addStretch()
        # Save folder
        save_button_exp = QHBoxLayout()
        save_button_exp.addWidget(self.widget_controls['label']['export_change_save'])
        save_button_exp.addWidget(self.widget_controls['button']['export_change_save'])
        save_button_exp.addStretch()
        export_controls.addLayout(save_button_exp)
        # export images button
        export_button = QHBoxLayout()
        export_button.addWidget(self.widget_controls['button']['export_images'])
        export_button.addStretch()
        export_controls.addLayout(export_button)

        ''' add to parameter controls tab'''
        self.tabs['slider'] = QTabWidget()
        for tab_layout, tab_name in zip([image_controls, metric_controls, settings_controls, experiment_controls, export_controls],
                                        ['transforms', 'metric params', 'image settings', 'experiment', 'export']):
            utils.add_layout_to_tab(self.tabs['slider'], tab_layout, tab_name)

        '''re calc graphs button'''
        graph_button = QVBoxLayout()
        if (self.metrics_avg_graph or self.metric_range_graph):
            graph_button.addWidget(self.widget_controls['button']['force_update'])
            graph_button.addWidget(self.widget_controls['check_box']['graph_limits'])
            graph_button.addStretch()

        ''' put the whole layout together '''
        image_left_side = QVBoxLayout()
        image_left_side.addLayout(image_layouts)
        image_left_side.addLayout(dataset_layout)
        image_middle = QVBoxLayout()
        image_middle.addLayout(metric_layouts)

        all_images = QHBoxLayout()
        all_images.addLayout(image_left_side)
        all_images.addLayout(image_middle)

        controls = QVBoxLayout()
        controls.addWidget(self.tabs['slider'])

        images_plus_controls = QVBoxLayout()
        images_plus_controls.addLayout(all_images)
        images_plus_controls.addLayout(controls)

        graph_right = QVBoxLayout()
        graph_right.addLayout(graph_layouts)
        graph_right.addLayout(graph_button)

        self.main_layout = QHBoxLayout()
        self.main_layout.addLayout(images_plus_controls)
        self.main_layout.addLayout(graph_right)

        # self.main_layout.setColumnStretch(1, 1)
        # self.main_layout.setSpacing(0)
        # self.main_layout.setContentsMargins(0, 0, 0, 0)

    def init_style(self, style='light', css_file=None):
        if css_file == None:
            dir = os.path.dirname(os.path.abspath(__file__))
            # css_file = os.path.join(dir, 'style-light.css')
            css_file = os.path.join(dir, f'style-{style}.css')
        if os.path.isfile(css_file):
            with open(css_file, 'r') as file:
                self.app.setStyleSheet(file.read())
        else:
            warnings.warn('Cannot load css style sheet - file not found')



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from IQM_Vis.transforms.effects import ( 
                         blur, 
                         brightness, 
                         binary_threshold, 
                         jpeg_compression, 
                         contrast,
                         hue,
                         saturation,
                         brightness_hsv)
from IQM_Vis.transforms.affine import (rotation, x_shift, y_shift, zoom_image)
from IQM_Vis.transforms.additive_noise import (salt_and_pepper_noise, Gaussian_noise, noise_hypersphere)

def get_all_transforms():
    '''
    Get all the available transformation/distortions alongside their recommended parameter ranges

    Returns:
        - all (dict): names, min, max, init_value and function for each transform
    '''
    all = {
        'brightness':{'min':-1.0, 'max':1.0, 'function':brightness},
        'contrast': {'min': 0.5, 'max': 2.5, 'init_value': 1.0, 'function': contrast},
        'hue': {'min': -0.5, 'max': 0.5, 'function': hue},
        'saturation': {'min': -0.5, 'max': 0.5, 'function': saturation},
        # 'brightness_hsv': {'min': -1.0, 'max': 1, 'function': brightness_hsv},
        'blur': {'init_value': 1, 'min': 1, 'max': 41, 'normalise': 'odd', 'function': blur},
        'Gaussian Noise': {'init_value': 0.0, 'min': 0.0, 'max': 0.5, 'function': Gaussian_noise()},
        'noise hypersphere': {'init_value': 0.0, 'min': 0.0, 'max': 5.0, 'function': noise_hypersphere()},
        'salt pepper': {'init_value': 0.0, 'min': 0.0, 'max': 0.05, 'function': salt_and_pepper_noise},
        'jpg comp.':{'init_value':101, 'min':1, 'max':101, 'function':jpeg_compression},
        'rotation':{'min':-180, 'max':180, 'function':rotation},
        'x_shift': {'min':-0.1, 'max':0.1, 'function':x_shift, 'init_value': 0.0},
        'y_shift': {'min':-0.1, 'max':0.1, 'function':y_shift, 'init_value': 0.0},
        'zoom':    {'min': 0.8, 'max':1.2, 'function':zoom_image, 'init_value': 1.0, 'num_values':21},
        # 'threshold':{'min':-40, 'max':40, 'function':binary_threshold},
    }
    return all



# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

from IQM_Vis.ui_wrapper import make_UI
from IQM_Vis.data_handlers.data_api import dataset_holder
from IQM_Vis.transforms import *
from IQM_Vis.metrics import *
from IQM_Vis import examples
from IQM_Vis import utils
from IQM_Vis.version import __version__

# Author info
__author__ = "Matt Clifford"
__email__ = "matt.clifford@bristol.ac.uk"
__license__ = "BSD 3-Clause"



'''
main entry point to initialise the UI
'''
# Author: Matt Clifford <matt.clifford@bristol.ac.uk>
# License: BSD 3-Clause License

import os
import time
from functools import partial
from PyQt6.QtWidgets import QApplication, QMessageBox, QFileDialog
from PyQt6.QtCore import QCoreApplication
import IQM_Vis
from IQM_Vis.UI import layout, widgets, images, ProgressBar

class make_app(widgets, layout, images):
    def __init__(self, 
                 app,
                 data_stores: list,
                 transformations: dict,
                 metrics_info_format='graph', # graph or text
                 metrics_avg_graph=False,
                 metric_range_graph=True,
                 metric_params: dict={},
                 image_display_size=300,
                 default_save_dir=IQM_Vis.utils.save_utils.DEFAULT_SAVE_DIR,
                 default_dataset_name='dataset1',
                 restrict_options=None,
                 num_steps_range=11,
                 num_step_experiment=6,
                 num_images_scroll_show=7,
                 test=False
                 ):
        super().__init__()
        self.app = app
        self.data_stores = data_stores
        self.transformations = transformations
        self.num_steps_range = num_steps_range
        self.num_step_experiment = num_step_experiment
        self.num_images_scroll_show = num_images_scroll_show
        self.preview_num = 0
        self.test = test

        self.metrics_info_format = metrics_info_format
        self.metrics_avg_graph = metrics_avg_graph
        self.metric_range_graph = metric_range_graph
        self.metric_params = metric_params
        self.default_save_dir = default_save_dir
        self.default_dataset_name = default_dataset_name
        self.restrict_options = restrict_options

        self.data_lims = {'fixed': 1, 'range_data': 1}
        self.plot_data_lim = 1

        self.dataset = self._single_image_or_dataset()
        self.setWindowTitle('IQM-Vis')

        self.make_status_bar()
        self.make_menu()

        self.widget_settings = {}
        self.image_display_size = image_display_size
        self.graph_size = 35
        self.rgb_brightness = 250  # max candela/m2 of rbg image
        self.display_brightness = 250  # max candela/m2 of display

        self.human_experiment_cache = {}
        self.construct_UI()

    def make_menu(self):
        self.menu_bar = self.menuBar()

        # make file menu
        self.file_menu = self.menu_bar.addMenu('File')
        save_dir_action = self.file_menu.addAction('Set Save Folder')
        load_image_action = self.file_menu.addAction('Load New Image')
        load_images_action = self.file_menu.addAction('Load New Image Folder')
        load_human_action = self.file_menu.addAction('Load Human Scores')
        load_experiment = self.file_menu.addAction('Load Experiment')
        reload_action = self.file_menu.addAction('Redo Graphs')
        quit_action = self.file_menu.addAction('Quit')

        save_dir_action.setStatusTip('Choose a folder to save project/experiments')
        save_dir_action.triggered.connect(self.change_save_folder)

        load_image_action.setStatusTip('Choose an image to load')
        load_image_action.triggered.connect(self.load_new_single_image)
        load_images_action.setStatusTip('Choose a folder of images to load')
        load_images_action.triggered.connect(self.load_new_images_folder)

        load_human_action.setStatusTip('Choose a human scores from csv file')
        load_human_action.triggered.connect(self.load_human_experiment)

        load_experiment.setStatusTip('Load experiment setup for folder')
        load_experiment.triggered.connect(self.load_experiment_from_dir)

        reload_action.setShortcut('Ctrl+R')
        reload_action.setStatusTip('')
        # reload_action.triggered.connect(self.display_images)
        reload_action.triggered.connect(partial(self.redo_plots, True))

        quit_action.setShortcut('Ctrl+Q')
        quit_action.setStatusTip('Exit application')
        quit_action.triggered.connect(self.quit)

        self.get_menu_checkboxes()

        self.edit_menu = self.menu_bar.addMenu('Load')
        load_transforms = self.edit_menu.addAction('All Transforms')
        load_metrics = self.edit_menu.addAction('All Metrics')
        load_metric_images = self.edit_menu.addAction('All Metric Images')

        load_transforms.triggered.connect(self.load_all_transforms)
        load_metrics.triggered.connect(self.load_all_metrics)
        load_metric_images.triggered.connect(self.load_all_metric_images)

        self.cache_menu = self.menu_bar.addMenu('Cache')
        clear_all_cache = self.cache_menu.addAction('Clear All Cache')

        clear_all_cache.triggered.connect(self.clear_all_cache_data)


    def load_all_transforms(self):
        all_trans_iqm_vis = IQM_Vis.transforms.get_all_transforms()
        for trans, data in all_trans_iqm_vis.items():
            if trans not in self.transformations:
                self.transformations[trans] = data
        self._remake_menu()
        self.construct_UI()

    def load_all_metrics(self):
        if not hasattr(self.data_stores[0], 'add_metric'):
            return
        all_metrics_iqm_vis = IQM_Vis.metrics.get_all_metrics()
        for metric, data in all_metrics_iqm_vis.items():
            self.data_stores[0].add_metric(metric, data)
        self._remake_menu()
        self.construct_UI()

    def load_all_metric_images(self):
        if not hasattr(self.data_stores[0], 'add_metric_image'):
            return
        all_metric_images_iqm_vis = IQM_Vis.metrics.get_all_metric_images()
        for metric, data in all_metric_images_iqm_vis.items():
            self.data_stores[0].add_metric_image(metric, data)
        self._remake_menu()
        self.construct_UI()

    def clear_all_cache_data(self):
        for data_store in self.data_stores:
            if hasattr(data_store, 'clear_all_cache'):
                data_store.clear_all_cache()
        self.status_bar.showMessage('Cleared all cache data', 8000)

    def quit(self):
        # QApplication.instance().quit()
        self.app.quit()

    def __del__(self):
        # garbage collection
        self.range_worker.stop()

    def closeEvent(self, event):
        # self.close_answer = None
        # if self.test == False:

        # Ask for confirmation
        self.close_answer = QMessageBox.question(self,
        "Confirm Exit...",
        "Are you sure you want to exit?\nAll unsaved data will be lost.",
        QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                                    QMessageBox.StandardButton.Yes)

        event.ignore()
        # if self.test == True or self.close_answer == QMessageBox.StandardButton.Yes:
        if self.close_answer == QMessageBox.StandardButton.Yes:
            if hasattr(self, 'range_worker'):
                self.range_worker.stop()

            if hasattr(self, 'range_worker_thread'):
                if self.range_worker_thread.isRunning():
                    self.range_worker_thread.quit()
                    self.range_worker_thread.wait()
            
            event.accept()
            # QCoreApplication.quit()

    def get_menu_checkboxes(self):
        ''' list all trans/metrics in the menu drop downs '''
        self.menu_options = {'transforms': {},
                             'metrics': {},
                             'metric_images': {}}
        # check if we need to restore current checked
        if hasattr(self, 'restore_checked_menu_items'):
            restrict_options = self.restore_checked_menu_items
        else:  # or building for the first time
            restrict_options = self.restrict_options
        
        # make menus
        set_checked_menu_from_iterable(self.menu_bar,
                                       self.transformations,
                                       'Transforms',
                                       self.menu_options['transforms'],
                                       self.construct_UI,
                                       restrict_options['transforms'])
        set_checked_menu_from_iterable(self.menu_bar,
                                       self.data_stores[0].metric_images,
                                       'Metric Images',
                                       self.menu_options['metric_images'],
                                       self.construct_UI,
                                       restrict_options['metric_images'])
        set_checked_menu_from_iterable(self.menu_bar,
                                       self.data_stores[0].metrics,
                                       'Metrics',
                                       self.menu_options['metrics'],
                                       self.construct_UI,
                                       restrict_options['metrics'])
    
    def _remake_menu(self):
        self._save_checked_menu_items()
        self.menu_bar.clear()
        self.make_menu()

    def _save_checked_menu_items(self):
        self.restore_checked_menu_items = {}
        for menu in ['transforms', 'metric_images', 'metrics']:
            self.restore_checked_menu_items[menu] = {}
            for item in self.menu_options[menu]:
                self.restore_checked_menu_items[menu][item] = self.menu_options[menu][item].isChecked()


    def make_status_bar(self):
        self.status_bar = self.statusBar()
        self.pbar = ProgressBar(self, minimum=0, maximum=100, textVisible=False,
                        objectName="GreenProgressBar")
        self.pbar.setValue(0)
        self.status_bar.addPermanentWidget(self.pbar)

    def construct_UI(self):
        if hasattr(self, 'range_worker'):
            self.range_worker.stop() # stop any calculations on the old UI
        # if hasattr(self, 'wait_until_safe_to_change_image'):
        #     self.wait_until_safe_to_change_image()
        # get currently selected tranformations to use
        self.checked_transformations = {}
        for trans, item in self.transformations.items():
            if self.menu_options['transforms'][trans].isChecked():
                self.checked_transformations[trans] = item
        # get currently selected metrics to use
        self.checked_metrics = []
        for metric in self.data_stores[0].metrics:
            if self.menu_options['metrics'][metric].isChecked():
                self.checked_metrics.append(metric)
        # get currently selected metric_images to use
        self.checked_metric_images = []
        for metric_image in self.data_stores[0].metric_images:
            if self.menu_options['metric_images'][metric_image].isChecked():
                self.checked_metric_images.append(metric_image)

        ''' update these '''
        # get any current tabs showing so we can keep them showing on a remake
        tabs_index = {}
        if hasattr(self, 'tabs'):
            tabs_index['slider'] = self.tabs['slider'].currentIndex()
            tabs_index['graph'] = self.tabs['graph'].currentIndex()
        else:
            tabs_index['slider'] = 0
            tabs_index['graph'] = 1

        # reset any range/correlation calc cache
        self.metric_range_graph_num = 0
        self.metric_correlation_graph_num = 0
        self.reset_correlation_data()

        # init the UI widgets and layouts
        self.init_style()     # layout.py
        self.init_widgets()   # widgets.py
        self.change_data(0, _redo_plots=True)   # images.py
        if self.dataset:
            self.set_preview_images(self.preview_num)
        self.main = self.init_layout()    # layout.py
        self.tabs['slider'].setCurrentIndex(tabs_index['slider'])
        self.tabs['graph'].setCurrentIndex(tabs_index['graph'])
        # self.experiments_tab.setCurrentIndex(experi_tabs_index)
        self.reset_sliders()  # widgets.py

        # self.setMinimumSize(self.main_widget.sizeHint())
        self.resize(self.main_widget.sizeHint())

    def reset_correlation_data(self):
        self.correlation_data = {}
        for i, _ in enumerate(self.data_stores):
            self.correlation_data[i] = {}

    def _single_image_or_dataset(self):
        '''set whether dataset or single image used for data_store'''
        for data_store in self.data_stores:
            try:
                data_store[1]
                dataset_found = True
                self.data_num = 0
                data_store[self.data_num]  # return to initial image
            except:
                return False

        # calc what is the last data point in all the datasets
        max_length = 0
        for data_store in self.data_stores:
            try:
                length = len(data_store)
                if length > max_length:
                    max_length = length
            except:
                obj_name = data_store.__class__.__name__
                raise AttributeError(f'{obj_name} needs to have __len__ attribute to be a dataset')
        self.max_data_ind = max_length - 1
        return dataset_found
    
    def change_save_folder(self):
        ''' change the save folder we are using '''
        # get the file opener for the user
        if os.path.exists(self.default_save_dir):
            start_dir = self.default_save_dir
        else:
            start_dir = os.path.expanduser("~")
        try:
            dir = QFileDialog.getExistingDirectory(self,
                                                   'Choose Save folder',
                                                   start_dir)
        except:
            return

        if dir == '':
            return
        
        self.default_save_dir = dir
        self.status_bar.showMessage(f'Changed save dir to: {self.default_save_dir}', 8000)
        self.widget_controls['label']['exp_change_save'].setText(f'Save Folder: {self.default_save_dir}')
        self.widget_controls['label']['export_change_save'].setText(f'Save Folder: {self.default_save_dir}')

def set_checked_menu_from_iterable(main_menu, iterable, name, action_store, connect_func, restrict_options=None):
    _menu = main_menu.addMenu(name)
    _menu.triggered.connect(connect_func)
    for i, trans in enumerate(iterable):
        action_store[trans] = _menu.addAction(trans)
        action_store[trans].setCheckable(True)
        if isinstance(restrict_options, int):
            if i  < restrict_options:
                action_store[trans].setChecked(True)
            else:
                action_store[trans].setChecked(False)
        elif isinstance(restrict_options, dict):
            if trans in restrict_options.keys():
                action_store[trans].setChecked(restrict_options[trans])
            else:
                action_store[trans].setChecked(False)
        else:
            action_store[trans].setChecked(True)
